from .const import DriverFunctions

import logging
from datetime import datetime, timedelta
from typing import Dict, TypedDict, Any

_LOGGER = logging.getLogger(__name__)

class DriverPingInfo:

    _mac_adress = None
    _driver_model = None

    def __init__(self):
        now = datetime.now()

        self._last_update = now.strftime("%H:%M")

    def update(self):
        now = datetime.now()

        self._last_update = now.strftime("%H:%M")

    def last_update(self) -> int:
        now = datetime.now()
        last = datetime.strptime(self._last_update, "%H:%M")
        last = datetime.combine(now.date(), last.time())

        if last > now:
            last -= timedelta(days=1)

        delta = now - last
        return int(delta.total_seconds() / 60)

    def set_driver_options(self, mac: str, model: str) -> None:
        self._mac_adress = int(mac)
        self._driver_model = int(model)

    def __repr__(self):
        return f"{self._last_update},{self._mac_adress},{self._driver_model}"

class GryfDriverFunctions():

    inputs = False
    shutters = False
    outputs = False
    temp = []
    pwms = []

class ObservableDict(dict):
    def __init__(self, name, ptr=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._name = name
        self._ptr = ptr

    def __setitem__(self, key, value):
        super().__setitem__(key, value)

        self._ptr(self._name, key, value)

class GryfData(Dict):

    _drivers = []

    _key_list = [
        DriverFunctions.OUTPUTS,
        DriverFunctions.INPUTS,
        DriverFunctions.PWM,
        DriverFunctions.PONG,
        DriverFunctions.TEMP,
        DriverFunctions.COVER,
    ]

    _subscription_list = {}

    def something_changed(self, key, key1, value):
        if key in [DriverFunctions.OUTPUTS, DriverFunctions.INPUTS, DriverFunctions.COVER]:
            if self._subscription_list.get(key1) == None:
                self._subscription_list[key1] = GryfDriverFunctions()

            _LOGGER.debug(key)

            if key == DriverFunctions.OUTPUTS:
                self._subscription_list[key1].outputs = True

            if key == DriverFunctions.COVER:
                self._subscription_list[key1].shutters = True

            if key == DriverFunctions.INPUTS:
                self._subscription_list[key1].inputs = True

    def __new__(cls):
        return super(GryfData, cls).__new__(cls)

    def __init__(self):
        super().__init__()
        for key in self._key_list:
            self[key] = ObservableDict(key, ptr=self.something_changed)
        self._subscription_list = {}

    def __repr__(self):
        return f"GryfData({dict.__repr__(self)})"

    def __getitem__(self, key):
        if key in self._key_list:
            return super().__getitem__(key)
        _LOGGER.error(f"Bad dict key: {key}")
        raise KeyError(f"Invalid key: {key}")

    def __setitem__(self, key, value):
        if key in self._key_list:
            return super().__setitem__(key, value)
        _LOGGER.error(f"Bad dict key: {key}")
        raise KeyError(f"Invalid key: {key}")

    def get(self, key, default=None):
        if key in self._key_list:
            return super().get(key, default)
        _LOGGER.error(f"Bad dict key in get(): {key}")
        return default

    def setdefault(self, key, default=None):
        if key in self._key_list:
            return super().setdefault(key, default)
        _LOGGER.error(f"Bad dict key in setdefault(): {key}")
        raise KeyError(f"Invalid key: {key}")

    def update(self, other=None, **kwargs):
        if other:
            for key in other:
                self[key] = other[key]
        for key in kwargs:
            self[key] = kwargs[key]

    def update_pong(self, id: int):
        if not self[DriverFunctions.PONG].get(id, None):
            self[DriverFunctions.PONG][id] = DriverPingInfo()

        self[DriverFunctions.PONG][id].update()

    def update_model(self, id: int, mac: str, model: str):
        if not self[DriverFunctions.PONG].get(id, None):
            self[DriverFunctions.PONG][id] = DriverPingInfo()

        self[DriverFunctions.PONG][id].set_driver_options(mac, model)
        
    @property
    def inputs(self):
        return self[DriverFunctions.INPUTS]

    @property
    def outputs(self):
        return self[DriverFunctions.OUTPUTS]

    @property
    def pwms(self):
        return self[DriverFunctions.PWM]

    @property
    def finds(self):
        return self[DriverFunctions.FIND]

    @property
    def pongs(self):
        return self[DriverFunctions.PONG]

    @property
    def temps(self):
        return self[DriverFunctions.TEMP]

    @property
    def covers(self):
        return self[DriverFunctions.COVER]



from pygryfsmart.const import DriverFunctions
from pygryfsmart import GryfApi

from datetime import datetime

class _GryfDevice:
    _name: str
    _id: int
    _pin: int
    _api: GryfApi
    _attributes = {}

    def __init__(self,
                 name: str,
                 id: int,
                 pin: int,
                 api: GryfApi
                 ) -> None:
        self._name = name
        self._id = id
        self._pin = pin
        self._api = api

        now = datetime.now()
        self._api.feedback.data[DriverFunctions.PONG][self._id] = now.strftime("%H:%M") 

    @property
    def available(self):
        return self._api.avaiable_module(self._id)

    @property
    def extra_attributes(self):
        return self._attributes



from pygryfsmart import GryfApi
from .base import _GryfDevice

class GryfInputLine(_GryfDevice):

    def __init__(self,
                 name: str,
                 api: GryfApi,
                 ) -> None:
        super().__init__(name, 
                         0, 
                         0, 
                         api)

        self._attributes = {}

    def subscribe(self , update_fun_ptr):
        self._api.subscribe_input_message(update_fun_ptr)

    @property
    def name(self):
        return f"{self._name}"




"""File with GryfApi driver functions."""

from .const import (
    DriverFunctions,
    OutputActions,
    DriverActions,
    KeyModes,
    ConfigurationFunctions,
    ShutterStates,
)
from .communications import _GryfCommunicationApiBase

import asyncio

class _GryfFunctionsApiBase(_GryfCommunicationApiBase):
    """GryfFunctionsApiBase."""
    
    def __init__(
            self,
            port: str,
            callback=None
        ) -> None:
        """Initialise GryfFunctionsApiBase."""

        super().__init__(port , callback=callback)

    async def set_out(
            self,
            id: int,
            pin: int,
            state: OutputActions | int
        ) -> None:
        """Set driver output state."""

        states = ["0"] * 8 if pin > 6 else ["0"] * 6
        states[pin - 1] = str(state)

        command = f"{DriverActions.SET_OUT}={id}," + ",".join(states) + "\n\r"
        await self.send_data(command)

    async def set_key_time(
            self,
            ps_time: int,
            pl_time: int,
            id: int,
            pin: int,
            type: KeyModes | int
        ) -> None:
        """Set target short and long press time for driver input."""

        command = f"{ConfigurationFunctions.SET_PRESS_TIME}={id},{pin},{ps_time},{pl_time},{type}\n\r"
        await self.send_data(command)

    async def ping_connection(self) -> bool:
        """Ping connection with drivers."""

        return await self.ping(1)

    async def search_module(
            self,
            id: int
        ) -> None:
        """Get model off current driver."""

        if id != 0:
            command = f"{DriverActions.SEARCH}=0,{id}\n\r"
            await self.send_data(command)
        else:
            await self.search_modules()

    async def search_modules(
            self,
            last_module: int | None=None
        ) -> None:
        """Get model off current drivers."""

        if last_module == None:
            module_count = max(len(self.feedback.data[DriverFunctions.INPUTS]), len(self.feedback.data[DriverFunctions.OUTPUTS]))
        else:
            module_count = last_module

        for i in range(module_count):
            command = f"{DriverActions.SEARCH}=0,{i + 1}\n\r"
            await self.send_data(command)

    async def ping(
            self,
            module_id: int
        ) -> bool:
        """Ping single driver."""
        
        command = f"{DriverActions.PING}={module_id}\n\r"
        await self.send_data(command)
        await asyncio.sleep(0.05)
        if self._last_ping == module_id:
            self._last_ping = 0
            return True
        self._last_ping = 0
        return False
    
    async def set_pwm(
            self,
            id: int,
            pin: int,
            level: int,
            extra: int=1,
        ) -> None:
        """Set current driver pwm output level."""

        if extra:
            command = f"{DriverActions.SET_PWM}={id},{pin},{level},1\n\r"
        else:
            command = f"{DriverActions.SET_PWM}={id},{pin},{level}\n\r"
        await self.send_data(command)

    async def set_cover(
            self,
            id: int,
            pin: int,
            time: int,
            operation: ShutterStates | int
        ) -> None:
        """Set current driver cover output state."""

        if operation in {ShutterStates.CLOSE , ShutterStates.OPEN , ShutterStates.STOP , ShutterStates.STEP_MODE} and pin in {1 , 2 , 3 , 4}:
            states = ["0"] * 4
            states[pin - 1] = str(operation)
            control_sum = id + time + int(states[0]) + int(states[1]) + int(states[2]) + int(states[3])

            command = f"{DriverActions.SET_COVER}={id},{time},{states[0]},{states[1]},{states[2]},{states[3]},{control_sum}\n\r"
            await self.send_data(command)
        else:
            raise ValueError(f"Argument out of scope: id: {id} , pin: {pin} , time: {time}, operation: {operation}")

    async def async_update_states(self):
        output_states = self.feedback.data.outputs

        for index, data in output_states.items():
            states = []
            for x, item in data.items():
                states.append(item)

            command = f"{DriverActions.SET_OUT}={index}," + ",".join(str(i) for i in states) + "\n\r"

            await self.send_data(command)
            await asyncio.sleep(0.5)

    async def reset(
            self,
            module_id: int,
            update_states: bool
        ) -> None:
        """Reset all drivers (module_id == 0) or single."""
        
        if module_id == 0:
            command = "AT+RST=0\n\r"
            await self.send_data(command)

            if update_states == True:
                await asyncio.sleep(3)

                await self.async_update_states()

    async def update_inputs_info(self, driver_num: int):

        command = f"{DriverActions.GET_IN_STATE}={driver_num}"

        await self.send_data(command)

    async def update_outputs_info(self, driver_num: int):

        command = f"{DriverActions.GET_OUT_STATE}={driver_num}"

        await self.send_data(command)

    async def update_shutters_info(self, driver_num: int):

        command = f"{DriverActions.GET_SHUTTER_STATE}={driver_num}"

        await self.send_data(command)



from pygryfsmart import GryfApi
from pygryfsmart.const import DriverFunctions

from .base import _GryfDevice

class GryfTemperature(_GryfDevice):

    def __init__(self,
                 name: str,
                 id: int,
                 pin: int,
                 api: GryfApi,
                 callback=None
                 ) -> None:
        self._attributes = {
            "id": id,
            "pin": pin,
        }

        super().__init__(name, 
                         id, 
                         pin, 
                         api)
        if callback is not None:
            self._api.subscribe(self._id , self._pin , DriverFunctions.TEMP , callback)

    def subscribe(self , update_fun_ptr):
        self._api.subscribe(self._id , self._pin, DriverFunctions.TEMP , update_fun_ptr)

    @property
    def name(self):
        return f"{self._name}"



from pygryfsmart.const import OutputActions, DriverFunctions
from pygryfsmart import GryfApi

from .base import _GryfDevice

class GryfLock(_GryfDevice):
  def __init__(
    self,
    name: str,
    id: int,
    pin: int,
    in_id: int | None,
    in_pin: int | None,
    api: GryfApi,
    update_fun_ptr=None
  ):
    super().__init__(
      name,
      id,
      pin,
      api
    )

    self._name = name
    self.name = name
    self._id = id
    self._pin = pin

    if in_id is not None:
      self._attributes = {
        "id_out": id,
        "pin_out": pin,
        "id_in": in_id,
        "pin_in": in_pin,
      }

      self._in_id = in_id
      self._in_pin = in_pin
      self._in_en = True

    else:
      self._attributes = {
        "id": id,
        "pin": pin,
      }
      self._in_en = False

    self._open = False
    self._locked = True
    self._update_fun_ptr = None

  async def async_update_in(self, state):
    self._open = not state

    data = {
      "open": self._open,
      "locked": self._locked,
    }

    await self._update_fun_ptr(data)

  async def update_out(self, state):
    self._locked = state
  
    data = {
      "open": self._open,
      "locked": self._locked,
    }

    await self._update_fun_ptr(data)

  def subscribe(self, update_fun_ptr):
    self._update_fun_ptr = update_fun_ptr

    self._api.subscribe(self._id, self._pin, DriverFunctions.OUTPUTS, self.update_out)
    if self._in_en:
      self._api.subscribe(self._in_id, self._in_pin, DriverFunctions.INPUTS, self.async_update_in)

  @property
  def output_enable(self):
    return self._in_en

  async def turn_on(self):
    await self._api.set_out(self._id, self._pin, OutputActions.ON)

  async def turn_off(self):
    await self._api.set_out(self._id, self._pin, OUTPUT_STATES.OFF)



from pygryfsmart import GryfApi
from pygryfsmart.const import DriverFunctions

from .base import _GryfDevice

import asyncio

class GryfPwm(_GryfDevice):

    _last_level = 70
    _is_on: bool

    def __init__(self,
                 name: str,
                 id: int,
                 pin: int,
                 api: GryfApi,
                 callback=None,
                 ) -> None:
        self._attributes = {
            "id": id,
            "pin": pin,
        }

        super().__init__(name, 
                         id, 
                         pin, 
                         api)
        self._last_level = 0
        self._is_on = False
        if callback is not None:
            self._api.subscribe(self._id , self._pin , DriverFunctions.PWM , callback)

    def subscribe(self , update_fun_ptr):
        self._api.subscribe(self._id , self._pin, DriverFunctions.PWM , update_fun_ptr)

    async def set_level(self , level: int):
        if level > 0:
            self._last_level = level
        await self._api.set_pwm(self._id , self._pin , level)
        await self._api.send_data(f"stateLED={self._id}\n\r")

    async def turn_on(self):
        await self._api.set_pwm(self._id , self._pin , self._last_level)
        self._is_on = True

    async def turn_off(self):
        await self._api.set_pwm(self._id , self._pin , 0)
        self._is_on = False

    async def toggle(self):
        if self._is_on:
            await self.turn_off()
        else:
            await self.turn_on()

    @property
    def name(self):
        return f"{self._name}"



from .const import (
    CONF_ID,
    CONF_PIN,
    CONF_PTR,
    CONF_FUNCTION,
    DriverFunctions,
)
from .parsing import Parser
from .typing import GryfData

import logging
import traceback

_LOGGER = logging.getLogger(__name__)

class Feedback:

    _parser: Parser
    _data = GryfData()

    def __init__(self , callback=None) -> None:
        self.callback = callback
        self._subscribers = []
        self._temp_subscribers = []
        self._parser = Parser(self)

    @property
    def data(self):
        return self._data

    async def handle_subscribtion(self , function: str, id=0):
        if id == 0:
            try:
                for sub in self._subscribers:
                    if function == sub[CONF_FUNCTION]:
                        await sub[CONF_PTR](self._data.get(function , {}).get(sub.get(CONF_ID) , {}).get(sub.get(CONF_PIN) , 0))
            except Exception as e:
                _LOGGER.error(f"Error subscriber 1: {e}")
        else:
            try:
                for sub in self._subscribers:
                    if function == sub[CONF_FUNCTION] and id == sub[CONF_ID]:
                        driver_states = self._data[function][sub.get(CONF_ID)]

                        await sub[CONF_PTR](driver_states[sub[CONF_PIN]])

            except Exception as e:
                _LOGGER.error(f"Error subscriber 2: {e} (type: {type(e)})")
                _LOGGER.error(traceback.format_exc())

    async def handle_temp_subscribtion(self , id: int , pin: int):
        for sub in self._temp_subscribers:
            if id == sub[CONF_ID] and pin == sub[CONF_PIN]:
                data = self._data.temps.get(id, {}).get(pin)
                await sub[CONF_PTR](data)


    async def input_data(self , line):
        if line == "??????????":
            return
        try:
            parts = line.split('=')
            parsed_states = parts[1].split(',')
            last_state = parsed_states[-1].split(';')
            parsed_states[-1] = last_state[0]

            COMMAND_MAPPER = {
                DriverFunctions.INPUTS: lambda states , line : self._parser.parse_metod_1(states , line , DriverFunctions.INPUTS),
                DriverFunctions.OUTPUTS: lambda states , line : self._parser.parse_metod_1(states , line , DriverFunctions.OUTPUTS),
                DriverFunctions.PRESS_SHORT: lambda states , line : self._parser.parse_metod_2(states , line , DriverFunctions.INPUTS , 2),
                DriverFunctions.PRESS_LONG: lambda states , line : self._parser.parse_metod_2(states , line , DriverFunctions.INPUTS , 3),
                DriverFunctions.TEMP: lambda states , line : self._parser.parse_temp(states , line),
                DriverFunctions.PWM: lambda states , line : self._parser.parse_metod_3(states , line , DriverFunctions.PWM),
                DriverFunctions.COVER: lambda states , line : self._parser.parse_cover(states , line , DriverFunctions.COVER),
                DriverFunctions.FIND: lambda states , line: self._parser.parse_find(states),
                DriverFunctions.PONG: lambda states , line: self._parser.parse_pong(states),
            }

            if str(parts[0]).upper() in COMMAND_MAPPER:
                await COMMAND_MAPPER[str(parts[0]).upper()](parsed_states , line)

            if self.callback:
                await self.callback() 

        except Exception as e:
            _LOGGER.error(f"ERROR parsing data: {e} (type: {type(e)})")
            _LOGGER.error(traceback.format_exc())

    def subscribe(self , conf: dict):
        self._subscribers.append(conf)

    def subscribe_temp(self , conf: dict):
        self._temp_subscribers.append(conf)



import asyncio
import logging

from pygryfsmart.const import SCHUTTER_STATES
from pygryfsmart.api import GryfApi

_LOGGER = logging.getLogger(__name__)

class GryfCover:
    def __init__(self,
                 device: GryfApi,
                 id: int,
                 pin: int,
                 time: int) -> None:
        self._pin = pin
        self._id = id
        self._device = device
        self._time = time
        self._current_cover_position = 0
        self._position_to_move = 0
        self._direction = 0

    def get_time(self , postion_to_move: int):
        return (postion_to_move * self._time) // 100

    def update_current_position(self):
        postion_moved = 100 / self._time
        if self._direction == SCHUTTER_STATES.CLOSE:
            self._current_cover_position -= postion_moved
        elif self._direction == SCHUTTER_STATES.OPEN:
            self._current_cover_position += postion_moved
        _LOGGER.debug("postion: " , self._current_cover_position)
        

    async def __cover_task(self):
        while (self._direction == SCHUTTER_STATES.OPEN and self._current_cover_position >= self._position_to_move) or (self._direction == SCHUTTER_STATES.CLOSE and self._current_cover_position <= self._position_to_move):
            await asyncio.sleep(1)   
            self.update_current_position()
        self.__cover_task.cancel()

    async def set_cover_position(self , position: int):

        self._position_to_move = position
            
        if self._current_cover_position > position:
            self._direction = SCHUTTER_STATES.CLOSE
            time = self.get_time(self._current_cover_position - position)
            await self._device.set_cover(self._id , self._pin , time , SCHUTTER_STATES.CLOSE)
        elif self._current_cover_position < position:
            self._direction = SCHUTTER_STATES.CLOSE
            time = self.get_time(position - self._current_cover_position)
            await self._device.set_cover(self._id , self._pin , time , SCHUTTER_STATES.OPEN)

        if not self.__cover_task:
            self.__cover_task
            



"""Main api class function."""

from .functions import _GryfFunctionsApiBase
from .gryf_expert import GryfExpert
from .const import(
    CONF_ID,
    CONF_PIN,
    CONF_PTR,
    CONF_FUNCTION,
    DriverFunctions,
    subscriptable_function,
)

import logging

_LOGGER = logging.getLogger(__name__)

class GryfApi(_GryfFunctionsApiBase):
    """Base pygryfsmart class."""

    _gryf_expert: GryfExpert
    _cover_toggle_metod: int=1
    
    def __init__(self,
                 port: str,
                 callback=None
        ) -> None:
        """Init pygryfsmart."""

        super().__init__(port , callback)

        self._gryf_expert = GryfExpert(self)

    def set_cover_toggle_metod(self, metod):
        self._cover_toggle_metod = metod

    async def start_gryf_expert(self) -> None: 
        """Start gryf expert server."""
        await self._gryf_expert.start_server()

    async def stop_gryf_expert(self) -> None:
        """Stop gryf expert server."""
        await self._gryf_expert.stop_server()

    def subscribe(
            self,
            id: int, 
            pin: int,
            func: str | DriverFunctions,
            ptr
        ) -> None:
        """Subscribe feedback from drivers."""

        if func in subscriptable_function:

            data = {
                CONF_ID: id,
                CONF_PIN: pin,
                CONF_FUNCTION: func,
                CONF_PTR: ptr
            }

            self.feedback.data[func]._ptr(func, id, pin)

            if func != DriverFunctions.TEMP:
                self.feedback.subscribe(data)
            else:
                self.feedback.subscribe_temp(data)
        else:
            _LOGGER.error(f"Bad function to subscribe: {func}")

    def realize_function_from_arguments(
        self,
        function: str,
        id: int,
        extra: int,
        state: int,
    ) -> None:
        pass
